from mmap_ninja import base
from mmap_ninja import numpy as np_ninja
from mmap_ninja.base import Wrapped
from mmap_ninja.string import StringsMmap
from mmap_ninja.ragged import RaggedMmap
from pathlib import Path
from typing import Union, Callable, Optional


def open_existing(out_dir: Union[str, Path], wrapper_fn: Optional[Callable] = None, mode="r+b"):
    out_dir = Path(out_dir)
    type_str = base._file_to_str(out_dir / "type.ninja")
    if type_str == "numpy":
        memmap = np_ninja.open_existing(out_dir)
        if wrapper_fn is not None:
            return Wrapped(memmap, wrapper_fn)
        return memmap
    if type_str == "ragged":
        return RaggedMmap(out_dir, wrapper_fn=wrapper_fn)
    if type_str == "string":
        memmap = StringsMmap(out_dir, mode=mode)
        if wrapper_fn is not None:
            return Wrapped(memmap, wrapper_fn)
        return memmap
    raise ValueError(f'Unknown type "{type_str}" while trying to open "{out_dir}" !')



from copy import copy
from pathlib import Path
from typing import Union, Sequence

import numpy as np

from mmap_ninja import numpy, base


def _np_shape_extend(shapes, arr):
    numpy.extend(shapes, arr)


def _ragged_shape_extend(shapes, arr):
    shapes.extend(arr)


class RaggedMmap:
    def __init__(
        self,
        out_dir: Union[str, Path],
        wrapper_fn=None,
        mode="r",
        starts_key="starts",
        ends_key="ends",
        shapes_key="shapes",
        flattened_shapes_key="flattened_shapes",
    ):
        out_dir = Path(out_dir)
        out_dir.mkdir(exist_ok=True)

        self.starts_key = starts_key
        self.ends_key = ends_key
        self.shapes_key = shapes_key
        self.flattened_shapes_key = flattened_shapes_key

        data_file = out_dir / "data"
        self.data_file = Path(data_file)

        self.out_dir = out_dir
        self.wrapper_fn = wrapper_fn
        self.mode = mode

        self.shapes_are_flat = None
        self.shapes_ctor = None
        self.memmap = None
        self.starts = None
        self.ends = None
        self.shapes = None
        self.flattened_shapes = None
        self.range = None
        self.n = None
        self.range = None
        self.shapes_extension_fn = None

        if (self.out_dir / "shapes_are_flat.ninja").exists():
            self._reload_fields()

    def _reload_fields(self):
        self.shapes_are_flat = bool(base._file_to_int(self.out_dir / "shapes_are_flat.ninja"))
        self.shapes_ctor = numpy.open_existing if self.shapes_are_flat else RaggedMmap
        self.memmap = numpy.open_existing(self.out_dir, mode=self.mode)
        self.starts = numpy.open_existing(self.out_dir / self.starts_key, mode=self.mode)
        self.ends = numpy.open_existing(self.out_dir / self.ends_key, mode=self.mode)
        self.shapes = self.shapes_ctor(self.out_dir / self.shapes_key, mode=self.mode)
        self.flattened_shapes = numpy.open_existing(self.out_dir / self.flattened_shapes_key, mode=self.mode)
        self.range = np.arange(len(self.starts), dtype=np.int64)
        self.n = len(self.shapes)
        self.range = np.arange(self.n)
        self.shapes_extension_fn = _np_shape_extend if self.shapes_are_flat else _ragged_shape_extend

    def get_multiple(self, item):
        indices = self.range[item]
        return [self.__getitem__(idx) for idx in indices]

    def set_multiple(self, item, value):
        for i, idx in enumerate(self.range[item]):
            self.set_single(idx, value[i])

    def set_single(self, idx, value):
        start = self.starts[idx]
        end = self.ends[idx]
        value = np.asarray(value)
        self.memmap[start:end] = value

    def __getitem__(self, item):
        if self.starts is None:
            raise IndexError(f"RaggedMmap is empty!")
        if np.isscalar(item):
            return self.get_single(item)
        return self.get_multiple(item)

    def __setitem__(self, key, value):
        if np.isscalar(key):
            return self.set_single(key, value)
        return self.set_multiple(key, value)

    def __len__(self):
        if self.starts is None:
            return 0
        return len(self.starts)

    def get_single(self, item):
        start = self.starts[item]
        end = self.ends[item]
        shape = self.shapes[item]
        res = self.memmap[start:end]
        if self.shapes_are_flat and shape[0] == 0:
            res = res.item()
        else:
            res = res.reshape(shape)
        if self.wrapper_fn is not None:
            res = self.wrapper_fn(copy(res))
        return res

    def append(self, array: np.ndarray):
        self.extend([array])

    def extend(self, arrays: Sequence[np.ndarray]):
        if self.starts is None:
            RaggedMmap.from_lists(self.out_dir, arrays)
            self._reload_fields()
            return
        numpy_bytes_slices = numpy._lists_of_ndarrays_to_bytes(arrays, self.memmap.dtype)
        numpy.extend(self.memmap, numpy_bytes_slices.buffer)
        end = self.ends[-1]
        numpy.extend(self.starts, end + numpy_bytes_slices.starts)
        numpy.extend(self.ends, end + numpy_bytes_slices.ends)
        numpy.extend(self.flattened_shapes, numpy_bytes_slices.flattened_shapes)
        self.shapes_extension_fn(self.shapes, numpy_bytes_slices.shapes)
        self._reload_fields()

    def __repr__(self):
        base_repr = super().__repr__()
        return f"{base_repr} of length: {len(self)}"

    @classmethod
    def from_lists(
        cls,
        out_dir: Union[str, Path],
        lists: Sequence[np.ndarray],
        dtype=None,
        mode="r+",
        wrapper_fn=None,
        starts_key="starts",
        ends_key="ends",
        shapes_key="shapes",
        flattened_shapes_key="flattened_shapes",
    ):
        out_dir = Path(out_dir)
        out_dir.mkdir(exist_ok=True)
        numpy_bytes_slices = numpy._lists_of_ndarrays_to_bytes(lists, dtype)
        numpy.from_ndarray(out_dir / starts_key, np.array(numpy_bytes_slices.starts, dtype=np.int64))
        numpy.from_ndarray(out_dir / ends_key, np.array(numpy_bytes_slices.ends, dtype=np.int64))
        shapes_are_flat = all([len(shape) == 1 for shape in numpy_bytes_slices.shapes])
        base._int_to_file(int(shapes_are_flat), out_dir / "shapes_are_flat.ninja")

        if shapes_are_flat:
            numpy.from_ndarray(out_dir / shapes_key, numpy_bytes_slices.shapes)
        else:
            RaggedMmap.from_lists(out_dir / shapes_key, numpy_bytes_slices.shapes)
        numpy.from_ndarray(
            out_dir / flattened_shapes_key, np.array(numpy_bytes_slices.flattened_shapes, dtype=np.int64)
        )
        numpy.from_ndarray(out_dir, np.array(numpy_bytes_slices.buffer))
        base._str_to_file("ragged", out_dir / "type.ninja")
        return cls(
            out_dir=out_dir,
            wrapper_fn=wrapper_fn,
            mode=mode,
            starts_key=starts_key,
            ends_key=ends_key,
            shapes_key=shapes_key,
            flattened_shapes_key=flattened_shapes_key,
        )

    @classmethod
    def from_generator(cls, out_dir: Union[str, Path], sample_generator, batch_size: int, verbose=False, **kwargs):
        return base.from_generator_base(
            out_dir=out_dir,
            sample_generator=sample_generator,
            batch_size=batch_size,
            verbose=verbose,
            batch_ctor=cls.from_lists,
            **kwargs,
        )



import json
import os
import stat

import pytest

from mmap_ninja import generic
from mmap_ninja.string import StringsMmap


def test_base_case(tmp_path):
    list_of_strings = ["Torba", "Boiler", "a", "zele pitka", "", "popo"]

    memmap = StringsMmap.from_strings(tmp_path / "strings_memmap", list_of_strings, verbose=True)
    for i, string in enumerate(list_of_strings):
        assert string == memmap[i]
    memmap[:2] = ["Korbo", "Moiler"]
    memmap[2] = "b"
    assert list_of_strings[:2] != memmap[:2]
    assert list_of_strings[2] != memmap[2]
    memmap.close()


def test_open_existing(tmp_path):
    list_of_strings = ["Torba", "Boiler", "a", "zele pitka", "", "popo"]

    memmap = StringsMmap.from_strings(tmp_path / "strings_memmap", list_of_strings, verbose=True)
    memmap = StringsMmap(tmp_path / "strings_memmap")
    for i, string in enumerate(list_of_strings):
        assert string == memmap[i]
    memmap[:2] = ["Korbo", "Moiler"]
    memmap[2] = "b"
    assert list_of_strings[:2] != memmap[:2]
    assert list_of_strings[2] != memmap[2]
    memmap.close()


def test_extend(tmp_path):
    list_of_strings = ["Torba", "Boiler", "a", "zele pitka", "", "popo"]

    memmap = StringsMmap.from_strings(tmp_path / "strings_memmap", list_of_strings)
    memmap.extend(["new", "new2", "uga dunga"])
    assert len(memmap) == 9
    assert memmap[-1] == "uga dunga"
    assert memmap[-2] == "new2"
    assert memmap[-3] == "new"


def test_append(tmp_path):
    list_of_strings = ["Torba", "Boiler", "a", "zele pitka", "", "popo"]

    memmap = StringsMmap.from_strings(tmp_path / "strings_memmap", list_of_strings)
    memmap.append("new")
    assert len(memmap) == 7
    assert memmap[-1] == "new"


def generate_strs(n):
    for i in range(n):
        yield str(i)


@pytest.mark.parametrize("n", [30, 3])
def test_from_generator(tmp_path, n):
    memmap = StringsMmap.from_generator(tmp_path / "strings_memmap", generate_strs(n), 4, verbose=True)
    for i in range(n):
        assert str(i) == memmap[i]
    generic.open_existing(tmp_path / "strings_memmap")


def test_open_empty(tmp_path):
    memmap = StringsMmap(tmp_path / "strings_memmap")
    assert len(memmap) == 0
    memmap.append("Something")
    assert len(memmap) == 1


def test_index_error(tmp_path):
    memmap = StringsMmap(tmp_path / "strings_memmap")
    print(memmap)
    with pytest.raises(IndexError):
        memmap[0]


def test_json_wrapper(tmp_path):
    memmap = StringsMmap(tmp_path / "strings_memmap")
    memmap.append(json.dumps({"asdasd": "as"}))

    memmap = generic.open_existing(tmp_path / "strings_memmap", wrapper_fn=json.loads)
    print(memmap[0])


def test_read_only(tmp_path):
    out_path = tmp_path / 'read_only_str_memmap'
    list_of_strings = ['This is the first test string...', '... and the second one.']
    StringsMmap.from_strings(out_dir=out_path, strings=list_of_strings)
    # Make all memmap files read-only for the user
    for p in out_path.glob(r'**/*'):
        if not p.is_file():
            continue
        os.chmod(p, stat.S_IRUSR)
    # Open in read-only mode
    memmap = StringsMmap(out_dir=out_path, mode='rb')
    for i, string in enumerate(list_of_strings):
        assert string == memmap[i]



import pytest

from mmap_ninja import base, generic


def test_open_throws_exception_if_type_is_unknown(tmp_path):
    out_dir = tmp_path / "mmap"
    out_dir.mkdir(exist_ok=True)
    base._str_to_file("asd", out_dir / "type.ninja")

    with pytest.raises(ValueError):
        generic.open_existing(out_dir)



import os
import stat

import numpy as np
import pytest

from mmap_ninja import generic
from mmap_ninja.ragged import RaggedMmap
from joblib import delayed, Parallel


def test_base_case(tmp_path):
    simple = list(range(4))
    mmap = RaggedMmap.from_lists(tmp_path / "simple", simple)
    for i in range(4):
        assert i == mmap[i]
    assert len(mmap) == 4

    print(generic.open_existing(tmp_path / "simple"))


def test_open_existing_case(tmp_path):
    simple = list(range(4))
    mmap = RaggedMmap.from_lists(tmp_path / "simple", simple)
    mmap = RaggedMmap(tmp_path / "simple")
    for i in range(4):
        assert i == mmap[i]


def test_np_case(tmp_path):
    simple = [np.array([11, 13, -1, 17]), np.array([2, 3, 4, 19]), np.array([90, 12])]
    mmap = RaggedMmap.from_lists(tmp_path / "simple", simple)
    for i in range(3):
        assert np.allclose(simple[i], mmap[i])


def test_set_np_case(tmp_path):
    simple = [np.array([11, 13, -1, 17]), np.array([2, 3, 4, 19]), np.array([90, 12])]
    mmap = RaggedMmap.from_lists(tmp_path / "simple", simple)
    changed = np.array([-1, -1, -1, -2])
    mmap[1] = changed
    for i in range(3):
        if i == 1:
            assert np.allclose(changed, mmap[i])
        else:
            assert np.allclose(simple[i], mmap[i])


def test_set_multiple_case(tmp_path):
    simple = [np.array([11, 13, -1, 17]), np.array([2, 3, 4, 19]), np.array([90, 12])]
    mmap = RaggedMmap.from_lists(tmp_path / "simple", simple)
    changed = np.array([-1, -1, -1, -2])
    mmap[:2] = [changed, changed]
    for i in range(3):
        if i < 2:
            assert np.allclose(changed, mmap[i])
        else:
            assert np.allclose(simple[i], mmap[i])


def test_get_multiple_case(tmp_path):
    simple = [np.array([11, 13, -1, 17]), np.array([2, 3, 4, 19]), np.array([90, 12])]
    mmap = RaggedMmap.from_lists(tmp_path / "simple", simple)
    result = mmap[:2]
    for i in range(2):
        assert np.allclose(result[i], simple[i])


def test_wrapper(tmp_path):
    simple = [np.array([11, 13, -1, 17]), np.array([2, 3, 4, 19]), np.array([90, 12])]
    mmap = RaggedMmap.from_lists(tmp_path / "simple", simple, wrapper_fn=lambda x: np.array(x, dtype=np.int8))
    assert mmap[0].dtype == np.int8


def test_extend(tmp_path):
    simple = [np.array([11, 13, -1, 17]), np.array([2, 3, 4, 19]), np.array([90, 12])]
    mmap = RaggedMmap.from_lists(tmp_path / "base", simple, wrapper_fn=lambda x: np.array(x, dtype=np.int16))
    extended = [np.array([123, -1]), np.array([-1, 0, 123, 92, 12])]
    mmap.extend(extended)
    assert np.allclose(mmap[3], extended[0])
    assert np.allclose(mmap[4], extended[1])
    new_arr = np.array([-1, -2, 628])
    mmap.append(new_arr)
    assert np.allclose(mmap[5], new_arr)


def generate_arrs(n):
    for i in range(n):
        yield np.ones(12) * i


@pytest.mark.parametrize("n", [30, 3])
def test_from_generator(tmp_path, n):
    memmap = RaggedMmap.from_generator(tmp_path / "strings_memmap", generate_arrs(n), 4, verbose=True)
    for i in range(n):
        assert np.allclose(np.ones(12) * i, memmap[i])


def test_nd_case(tmp_path):
    simple = [
        np.array([[11, 13], [-1, 17]]),
        np.array([[2, 3], [4, 19]]),
        np.array([[90], [12]]),
    ]
    mmap = RaggedMmap.from_lists(tmp_path / "base", simple, wrapper_fn=lambda x: np.array(x, dtype=np.int16))
    assert np.allclose(mmap[-1], simple[-1])


def test_different_number_of_axes(tmp_path):
    simple = [np.array([[11, 13], [-1, 17]]), np.array([2, 3]), np.array([[90], [12]])]
    mmap = RaggedMmap.from_lists(tmp_path / "base", simple, wrapper_fn=lambda x: np.array(x, dtype=np.int16))
    assert np.allclose(mmap[-1], simple[-1])


@pytest.fixture
def np_array_with_different_number_of_axes():
    return [np.array([[11, 13], [-1, 17]]), np.array([2, 3]), np.array([[90], [12]])]


def test_different_number_of_axes_gen(tmp_path, np_array_with_different_number_of_axes):
    mmap = RaggedMmap.from_generator(tmp_path / "base", np_array_with_different_number_of_axes, batch_size=1)
    assert np.allclose(mmap[-1], np_array_with_different_number_of_axes[-1])


def test_if_batch_size_exceeds_n_samples_crash(tmp_path, np_array_with_different_number_of_axes):
    mmap = RaggedMmap.from_generator(tmp_path / "base", np_array_with_different_number_of_axes, batch_size=100)
    assert np.allclose(mmap[-1], np_array_with_different_number_of_axes[-1])


def read_something(arr, i):
    return arr[i]


def test_parallel_read(tmp_path, np_array_with_different_number_of_axes):
    n_workers = 32
    RaggedMmap.from_generator(tmp_path / "base", np_array_with_different_number_of_axes, batch_size=100)
    ragged = RaggedMmap(tmp_path / "base")
    delayed_funcs = [delayed(read_something)(ragged, np.random.randint(len(ragged))) for _ in range(n_workers)]
    p = Parallel()
    r = p(delayed_funcs)
    assert len(r) == n_workers


def test_empty_init(tmp_path):
    ragged = RaggedMmap(tmp_path / "samples")
    assert len(ragged) == 0
    ragged.append(np.array([[1.0, 2.0, 3], [4.0, 5.0, 6.0]]))
    assert len(ragged) == 1


def test_ragged_index_error(tmp_path):
    ragged = RaggedMmap(tmp_path / "samples")
    with pytest.raises(IndexError):
        ragged[0]


def test_read_only(tmp_path, np_array_with_different_number_of_axes):
    out_path = tmp_path / 'read_only_ragged_memmap'
    RaggedMmap.from_generator(out_path, np_array_with_different_number_of_axes, batch_size=100)
    # Make all memmap files read-only for the user
    for p in out_path.glob(r'**/*'):
        if not p.is_file():
            continue
        os.chmod(p, stat.S_IRUSR)
    # Open in read-only mode
    ragged = RaggedMmap(out_path, mode='r')
    for i, arr in enumerate(np_array_with_different_number_of_axes):
        assert np.allclose(arr, ragged[i])



import numpy as np

from mmap_ninja import base


def test_int_conversions():
    a = 17
    assert a == base._bytes_to_int(base._int_to_bytes(a))


def test_int_file_conversions(tmp_path):
    a = 17
    file = tmp_path / "int"
    base._int_to_file(a, file)
    assert a == base._file_to_int(file)


def test_large_num():
    base._int_to_bytes(6653643750000, fmt="<Q")


def test_str_conversions():
    a = "ugabuga"
    assert a == base._bytes_to_str(base._str_to_bytes(a))


def test_str_file_conversions(tmp_path):
    a = "asd"
    file = tmp_path / "str"
    base._str_to_file(a, file)
    assert a == base._file_to_str(file)


def test_shape_conversions():
    shape = (1, 9, 10, 1)
    assert shape == base._bytes_to_shape(base._shape_to_bytes(shape))


def test_shape_file(tmp_path):
    shape = (1, 4, 5)
    file = tmp_path / "shape"
    base._shape_to_file(shape, file)
    assert shape == base._file_to_shape(file)


def test_wrapper():
    arr = np.array([1, 2, 3])
    wrapped = base.Wrapped(arr, lambda x: -x)
    assert wrapped[0] == -1
    assert wrapped[1] == -2
    assert wrapped[2] == -3
    assert len(wrapped) == 3



import sys
import os
from collections import defaultdict
from pathlib import Path

import numpy as np

from time import time

import pandas as pd

import matplotlib.pyplot as plt

from mmap_ninja.ragged import RaggedMmap

path = Path(sys.argv[1])
data_path = Path(sys.argv[2])


# https://stackoverflow.com/a/1392549/6636290
def get_dir_size(start_path):
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size


# Modified https://stackoverflow.com/a/31631711/6636290
def humanbytes(B):
    """Return the given bytes as a human friendly KB, MB, GB, or TB string."""
    B = float(B)
    KB = float(1024)
    GB = float(KB ** 3)  # 1,073,741,824
    return '{0:.6f} GB'.format(B / GB)


def main():
    df_data = defaultdict(list)

    start_t = time()
    in_memory = np.load(data_path / 'in_memory.npy', allow_pickle=True)
    df_data['in_memory'].append(time() - start_t)
    start_t = time()
    for i in range(len(in_memory)):
        img = in_memory[i]
    df_data['in_memory'].append(time() - start_t)
    df_data['in_memory'].append(humanbytes(sum([arr.nbytes for arr in in_memory])))
    df_data['in_memory'].append(humanbytes(os.path.getsize(data_path / 'in_memory.npy')))

    start_t = time()
    training_images = RaggedMmap(data_path / 'training_images')
    df_data['ragged_mmap'].append(time() - start_t)
    start_t = time()
    for i in range(len(training_images)):
        img = training_images[i]
    df_data['ragged_mmap'].append(time() - start_t)
    biggest_image_nbytes = in_memory[training_images.flattened_shapes.argmax()].nbytes
    df_data['ragged_mmap'].append(humanbytes(biggest_image_nbytes))
    df_data['ragged_mmap'].append(humanbytes(get_dir_size(data_path / 'training_images')))

    df_data['imread_from_disk'].append(0.)
    start_t = time()
    for file_path in path.iterdir():
        img = plt.imread(file_path)
    df_data['imread_from_disk'].append(time() - start_t)
    df_data['imread_from_disk'].append(humanbytes(biggest_image_nbytes))
    df_data['imread_from_disk'].append(humanbytes(get_dir_size(path)))

    benchmark = pd.DataFrame(df_data).T
    benchmark.columns = ['Initial load (s)', 'Time for iteration (s)', 'Memory usage (GB)', 'Disk usage (GB)']

    print()

    print(benchmark.to_markdown(floatfmt=".6f"))


if __name__ == '__main__':
    main()



import mmap
from pathlib import Path
from typing import Sequence, Union

import numpy as np

from mmap_ninja import numpy, base
from mmap_ninja.base import _bytes_to_str, _str_to_bytes, _sequence_of_strings_to_bytes


class StringsMmap:
    def __init__(
        self,
        out_dir: Union[str, Path],
        mode="r+b",
        starts_key="starts",
        ends_key="ends",
    ):
        out_dir = Path(out_dir)
        out_dir.mkdir(exist_ok=True)
        data_file = out_dir / "data.ninja"
        self.data_file = Path(data_file)
        self.mode = mode
        self.out_dir = out_dir
        self.starts_key = starts_key
        self.ends_key = ends_key

        self.starts = None
        self.ends = None
        self.range = None
        self.file = None
        self.buffer = None

        if (self.out_dir / self.starts_key / "dtype.ninja").exists():
            self._reload_fields()

    def get_multiple(self, item):
        indices = self.range[item]
        return [self.__getitem__(idx) for idx in indices]

    def get_single(self, item):
        start = self.starts[item]
        end = self.ends[item]
        return _bytes_to_str(self.buffer[start:end])

    def __getitem__(self, item):
        if self.starts is None:
            raise IndexError(f"StringsMmap is empty!")
        if np.isscalar(item):
            return self.get_single(item)
        return self.get_multiple(item)

    def __setitem__(self, key, value):
        if np.isscalar(key):
            return self.set_single(key, value)
        return self.set_multiple(key, value)

    def __len__(self):
        if self.starts is None:
            return 0
        return len(self.starts)

    def set_multiple(self, key, value):
        for i, idx in enumerate(self.range[key]):
            new_value: str = value[i]
            self.set_single(idx, new_value)

    def set_single(self, idx, new_value):
        start = self.starts[idx]
        end = self.ends[idx]
        self.buffer[start:end] = _str_to_bytes(new_value)

    def close(self):
        self.buffer.close()
        self.file.close()

    def extend(self, list_of_strings: Sequence[str], verbose=False):
        if self.starts is None:
            StringsMmap.from_strings(self.out_dir, list_of_strings, verbose=verbose)
            self._reload_fields()
            return
        bytes_slices = _sequence_of_strings_to_bytes(list_of_strings, verbose=verbose)
        end = self.ends[-1]
        start_offsets = end + bytes_slices.starts
        end_offsets = end + bytes_slices.ends
        numpy.extend(self.starts, start_offsets)
        numpy.extend(self.ends, end_offsets)
        self.close()
        out_dir = self.data_file.parent
        with open(out_dir / "data.ninja", "ab") as data_file:
            data_file.write(bytes_slices.buffer)
            data_file.flush()
        self._reload_fields()

    def _reload_fields(self):
        self.starts = numpy.open_existing(self.out_dir / self.starts_key, mode="r")
        self.ends = numpy.open_existing(self.out_dir / self.ends_key, mode="r")
        self.range = np.arange(len(self.starts), dtype=np.int64)
        self.file = open(self.data_file, mode=self.mode)
        access = mmap.ACCESS_READ if self.mode == 'rb' else mmap.ACCESS_DEFAULT
        self.buffer = mmap.mmap(self.file.fileno(), 0, access=access)

    def append(self, string: str):
        self.extend([string])

    def __repr__(self):
        base_repr = super().__repr__()
        return f"{base_repr} of length: {len(self)}"

    @classmethod
    def from_strings(
        cls,
        out_dir: Union[str, Path],
        strings: Sequence[str],
        mode="r+b",
        starts_key="starts",
        ends_key="ends",
        verbose=False,
    ):
        out_dir = Path(out_dir)
        out_dir.mkdir(exist_ok=True)
        bytes_slices = _sequence_of_strings_to_bytes(strings, verbose=verbose)
        with open(out_dir / "data.ninja", "wb") as f:
            f.write(bytes_slices.buffer)
        base._str_to_file("string", out_dir / "type.ninja")
        numpy.from_ndarray(out_dir / starts_key, np.array(bytes_slices.starts, dtype=np.int64))
        numpy.from_ndarray(out_dir / ends_key, np.array(bytes_slices.ends, dtype=np.int64))
        return cls(out_dir, mode=mode, starts_key=starts_key, ends_key=ends_key)

    @classmethod
    def from_generator(cls, out_dir: Union[str, Path], sample_generator, batch_size: int, verbose=False, **kwargs):
        return base.from_generator_base(
            out_dir=out_dir,
            sample_generator=sample_generator,
            batch_size=batch_size,
            verbose=verbose,
            batch_ctor=cls.from_strings,
            **kwargs,
        )



from dataclasses import dataclass
from pathlib import Path
from typing import Union, List, Tuple, Sequence, Optional, Dict

# See: https://numpy.org/doc/stable/reference/generated/numpy.memmap.html
import numpy as np
from mmap_ninja.base import from_generator_base

from mmap_ninja import base


def _create_if_not_exists(out_dir: Union[str, Path]) -> Path:
    """
    A helper mkdir that creates the directory if it doesn't exist.
    Returns the path to the directory.created.
    """
    out_dir = Path(out_dir)
    out_dir.mkdir(exist_ok=True, parents=True)
    return out_dir


def _save_mmap_kwargs(out_dir: Path, dtype: Union[np.dtype, str], shape: Sequence[int], order: str) -> None:
    """
    Persists the arguments needed for initializing a ``np.memmap``, so that the user does not have to specify them.

    :param out_dir: The directory in which the memory map is persisted
    :param dtype: The ``np.dtype`` of the memory map, represented either by a string or directly by its type.
    :param shape: A sequence of integers representing the shape of the memory map.
    :param order: A string, representing the order - ``"F"`` or ``"C"``
    :return:
    """
    out_dir = _create_if_not_exists(out_dir)
    base._str_to_file("numpy", out_dir / "type.ninja")
    base._str_to_file(np.dtype(dtype).name, out_dir / f"dtype.ninja")
    base._shape_to_file(shape, out_dir / f"shape.ninja")
    base._str_to_file(order, out_dir / f"order.ninja")


def _read_mmap_kwargs(out_dir: Path) -> Dict:
    """
    Reads already persisted arguments needed for opening a ``np.memmap``.

    :param out_dir: The persistence directory of the ``np.memmap``.
    :return: A dictionary representing the ``kwargs`` needed for initialization.
    """
    out_dir = _create_if_not_exists(out_dir)
    return {
        "dtype": base._file_to_str(out_dir / "dtype.ninja"),
        "shape": base._file_to_shape(out_dir / "shape.ninja"),
        "order": base._file_to_str(out_dir / "order.ninja"),
    }


def _empty(
    out_dir: Union[str, Path],
    dtype: Union[str, np.dtype],
    shape: Sequence[int],
    order: str,
) -> np.memmap:
    """
    Creates an empty ``np.memmap``

    :param out_dir: The persistence directory
    :param dtype: The dtype of the array
    :param shape: The shape of the array
    :param order: The order, either ``"C"`` or ``"F"``
    :return:
    """
    out_dir = _create_if_not_exists(out_dir)
    _save_mmap_kwargs(out_dir, dtype, shape, order)
    memmap = np.memmap(str(out_dir / "data.ninja"), mode="w+", dtype=dtype, shape=shape, order=order)
    return memmap


def from_ndarray(out_dir: Union[str, Path], arr: np.ndarray) -> np.memmap:
    """
    Initializes a memory map, in which all samples should be of the same shape

    :param out_dir: The directory in which the memory map will be persisted
    :param arr: The numpy array which should be memory mapped.
    :return: The memory mapped file
    """
    arr = np.asarray(arr)
    out_dir = _create_if_not_exists(out_dir)
    dtype = arr.dtype
    shape = arr.shape
    order = "F" if np.isfortran(arr) else "C"
    memmap = np.memmap(str(out_dir / "data.ninja"), mode="w+", dtype=dtype, shape=shape, order=order)
    memmap[:] = arr
    _save_mmap_kwargs(out_dir, dtype, shape, order)
    return memmap


def from_generator(out_dir: Union[str, Path], sample_generator, batch_size: int, n: int, verbose=False) -> np.memmap:
    """
    Create a numpy memory-map from a sample generator.

    :param sample_generator: A generator of the samples
    :param out_dir: The output directory
    :param batch_size: How often to flush to disk
    :param n: Total number of samples.
    :param verbose: Whether to show the progress bar.
    :return:
    """
    return from_generator_base(
        out_dir=out_dir,
        sample_generator=sample_generator,
        batch_size=batch_size,
        batch_ctor=from_ndarray,
        extend_fn=extend,
        verbose=verbose,
    )


def open_existing(out_dir: Union[str, Path], mode="r") -> np.memmap:
    """
    Open an already existing numpy array.

    :param out_dir: The output directory.
    :param mode: The mode with which to open the memory-mapped file.
    :return: The ``np.memmap`` object.
    """
    out_dir = Path(out_dir)
    kwargs = _read_mmap_kwargs(out_dir)
    memmap = np.memmap(str(out_dir / "data.ninja"), mode=mode, **kwargs)
    return memmap


def extend_dir(out_dir: Union[str, Path], arr: np.ndarray) -> None:
    """
    Extend an already existing memory map by adding new samples

    :param out_dir: The directory, in which the memory-mapped array is stored.
    :param arr: The numpy array of new samples
    :return:
    """
    arr = np.asarray(arr)
    out_dir = Path(out_dir)
    kwargs = _read_mmap_kwargs(out_dir)
    shape = kwargs["shape"]
    order = kwargs["order"]
    dtype = np.dtype(kwargs["dtype"])
    assert shape[1:] == arr.shape[1:], (
        f"Trying to append samples with incorrect shape: {arr.shape[1:]}, " f"expected: {shape[1:]}"
    )
    with open(out_dir / "data.ninja", "ab") as data_file:
        data_file.write(arr.astype(dtype).tobytes(order=order))
        data_file.flush()
    new_shape = (shape[0] + arr.shape[0], *shape[1:])
    base._shape_to_file(new_shape, out_dir / f"shape.ninja")


def extend(np_mmap: np.memmap, arr: np.ndarray) -> None:
    """
    Extend a numpy memory map with new samples

    :param np_mmap: The numpy memory map object
    :param arr: The new samples
    :return:
    """
    extend_dir(Path(np_mmap.filename).parent, arr)


def append(np_mmap: np.memmap, arr: np.ndarray) -> None:
    """
    Append a single sample to an already existing numpy array

    :param np_mmap:
    :param arr:
    """
    extend(np_mmap, np.expand_dims(np.asarray(arr), axis=0))


@dataclass
class NumpyBytesSlices:
    buffer: np.ndarray
    starts: List[int]
    ends: List[int]
    flattened_shapes: List[int]
    shapes: List[Tuple[int]]


def _lists_of_ndarrays_to_bytes(lists: Sequence[np.ndarray], dtype) -> NumpyBytesSlices:
    """
    Converts a list of numpy arrays into bytes.

    :param lists: The samples that have to be converted.
    :param dtype: The dtype of the arrays. If not provided, the dtype of the first sample will be used.
    :return: The numpy arrays as bytes
    """
    offset = 0
    starts = []
    ends = []
    shapes = []
    arrs = []
    flattened_shapes = []
    for l in lists:
        if dtype is None:
            arr = np.asarray(l)
            dtype = arr.dtype
        else:
            arr = np.asarray(l, dtype=dtype)
        flattened = arr.ravel()
        starts.append(offset)
        ends.append(offset + len(flattened))
        flattened_shapes.append(len(flattened))
        shapes.append(arr.shape if len(arr.shape) > 0 else (0,))
        arrs.append(flattened)
        offset += len(flattened)
    buffer = np.concatenate(arrs)
    return NumpyBytesSlices(buffer, starts, ends, flattened_shapes, shapes)



from setuptools import setup, find_packages
import pathlib
import mmap_ninja

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="mmap_ninja",
    version=mmap_ninja.__version__,
    description="mmap.ninja: Memory mapped data structures",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/hristo-vrigazov/mmap.ninja",
    author="Hristo Vrigazov",
    author_email="hvrigazov@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=("tests", "benchmark_images.py")),
    include_package_data=True,
    install_requires=["numpy", "tqdm"],
)
